# Frostbite Data Dumping Script by Frankelstner
# Updated and optimized for newer Frostbite titles by NoFaTe

# NOTE: This version has been adapted to work with FIFA 17
# and has only been tested with the FIFA 17 Demo.

import dbo
import os
import ctypes
import json
import zstd
import sys

from binascii import hexlify, unhexlify
from struct import pack, unpack
from cStringIO import StringIO
from ctypes import *

# Adjust paths here.
g_GameDirectory = r"M:\Games\FIFA 17 DEMO"
g_TargetDirectory = r"H:\Research\FIFA17\Dump\Files"

# You probably don't need to change these.
g_MainDataRoot = r"Data"

# This table covers all currently known resource types.
# New game or engine versions might introduce new types.
g_ResourceTypes = {
    0xC6DBEE07: ".AnimatedPointCloud",
    0xD070EED1: ".AnimTrackData",
    0x51A3C853: ".AssetBank",    
    0xE565EB15: ".DxShaderDatabase",
    0x10F0E5A1: ".DxShaderProgramDatabase",
    0x5C4954A6: ".DxTexture",
    0x70C5CB3E: ".EnlightenDatabase",
    0xE156AF73: ".EnlightenProbeSet",
    0x59CEEB57: ".EnlightenShaderDatabase",
    0x5BDFDEFE: ".EnlightenSystem",
    0x4864737B: ".HavokDestructionPhysicsData",
    0x91043F65: ".HavokPhysicsData",
    0xC78B9D9D: ".ImpulseResponse",
    0x36F3F2C0: ".IShaderDatabase",
    0xC417BBD3: ".ITexture",
    0x49B156D4: ".MeshSet",
    0x30B4A553: ".OccluderMesh",
    0x319D8CD0: ".RagdollResource",
    0x3568E2B7: ".RawFileData",
    0x41D57E10: ".RenderTexture",
    0x7AEFC446: ".StaticEnlightenDatabase",
    0x2D47A5FF: ".SwfMovie",
    0x6BB6D7D2: ".Terrain",
    0x15E1F32E: ".TerrainDecals",
    0x22FE8AC8: ".TerrainStreamingTree",
    0x1CA38E06: ".VisualTerrain",

    # Warsaw Additions
    0x957C32B1: ".AtlasTexture",
    0xAFECB022: ".CompiledLuaResource",
    0xBCC7FB86: ".Dx11Texture",
    0xF04F0C81: ".Dx11ShaderProgramDatabase",
    0xC6CD3286: ".EnlightenStaticDatabase",
    0xE36F0D59: ".HavokClothPhysicsData",
    0x31E779A2: ".MovieTexture",
    0xA23E75DB: ".TerrainLayerCombinations",

    # Whiteshark Additions
    0x6BDE20BA: ".Dx12Texture",
    0x9C4FAA17: ".HeightfieldDecal",
    0xC611F34A: ".MeshEmitterResource",
    0x9D00966A: ".UITtfFontFile",

    # Pamplona Additions
    0x387CA0AD: ".EAClothData",
    0xEFC70728: ".ZoneStreamerGrid",
    0x86521D6C: ".LinearMediaAsset",
    0x0DEAFE10: ".IesResource",
    0xC664A660: ".PamReplayResource",

    # FIFA Additions
    0x85EA8656: ".EAClothEntityData", 
    0x85AC783D: ".EAClothAssetData",
    0x3B9D1688: ".PSDResource",
    0x428EC9D4: ".BundleRefTableResource",
    0xEF23407C: ".FifaPhysicsResourceData",
    0x1091C8C5: ".MorphTargetsResource",
    0xB2C465F6: ".NewWaveResource",
}


def hex2(num): 
    return hexlify(pack(">i", num))


# Generic struct for the cat entry
class Stub(): 
    pass 


def fixguid(guid = ''):
    g = guid.strip()

    if len(g) != 32:
        return g

    fixedguid = g[6:8] + g[4:6] + g[2:4] + g[0:2] + '-'
    fixedguid += g[10:12] + g[8:10] + '-'
    fixedguid += g[14:16] + g[12:14] + '-'
    fixedguid += g[16:20] + '-'
    fixedguid += g[20:]

    return fixedguid.upper()


def readCat(catDict, catPath, rootPath):
    entryCount = 0
    entryCounter = 0

    cat = dbo.UnXor(catPath)[0]
    cat.seek(16) # NyanNyanNyanNyan

    entryCount = unpack("<Q",cat.read(8))[0]

    # Not sure where this came from. Maybe align-32?
    cat.read(16)

    casDirectory = os.path.dirname(catPath) + "\\"

    while entryCounter < entryCount:
        entry = Stub()
        sha1 = cat.read(20)
        entry.offset, entry.size, unk01, casNum = unpack("<IIII", cat.read(16))
        entry.path = casDirectory + "cas_" + ("0" + str(casNum) if casNum < 10 else str(casNum)) + ".cas"

        # TODO: Figure out what unk01 is at a later time.

        catDict[sha1] = entry
        entryCounter = entryCounter + 1

    print "Parsed %d entries in the catalog file '%s'." % (entryCounter, catPath.replace(rootPath, '')[1:])


def dump(tocPath, targetFolder):
    print "Dumping '%s'..." % tocPath

    toc = dbo.FromFile(tocPath)

    if not (toc.get("bundles") or toc.get("chunks")): 
        return

    sbPath = tocPath[:-3] + "sb"
    sb = open(sbPath, "rb") 

    for tocEntry in toc.bundles:
        print "Dumping bundle '%s' (%d)." % (tocEntry.id, tocEntry.offset)

        if tocEntry.get("base"): 
            continue
        
        sb.seek(tocEntry.offset)

        if toc.get("cas"):
            bundle = dbo.FromStream(sb, tocEntry.size)

            for listType in ("ebx", "res", "chunks"):
                if listType not in vars(bundle):
                    vars(bundle)[listType] = []
                
            for chunk in bundle.chunks:
                chunk.originalSize = chunk.logicalOffset + chunk.logicalSize
                    
            if tocEntry.get("delta"):
                raise Exception("Patch data is currently unsupported.")
            else:
                sourcePath = None 
        else:
            raise Exception("Non-CAS payloads are currently unsupported.")

        for entry in bundle.ebx:
            targetPath = targetFolder + "/bundles/ebx/" + entry.name + ".ebx"
            casPayload(entry, targetPath, sourcePath, entry.size)

        for entry in bundle.res:
            targetPath = targetFolder + "/bundles/res/" + entry.name + " " + hexlify(pack(">q",entry.resRid))

            if entry.resMeta != "\0" * 16: 
                targetPath += " " + hexlify(entry.resMeta)

            # Small hack because Python is "awesome".
            resType = ctypes.c_ulong(entry.resType).value

            if resType not in g_ResourceTypes: 
                print "Unknown resource type: %d for resource '%s'" % (entry.resType, targetPath)
                targetPath += ".unknownres " + hex2(entry.resType)
            else: 
                targetPath += g_ResourceTypes[resType]
            
            casPayload(entry, targetPath, sourcePath, entry.size)

        for i in xrange(len(bundle.chunks)):
            entry = bundle.chunks[i]
            targetPath = targetFolder + "/bundles/chunks/" + fixguid(hexlify(entry.id))
            casPayload(entry, targetPath, sourcePath, entry.size) 

    for entry in toc.chunks:
        targetPath = targetFolder + "/chunks/" + fixguid(hexlify(entry.id))

        if toc.get("cas"):
            catEntry = None

            try:
                catEntry = cat[entry.sha1]
            except:
                continue

            casPayload(entry, targetPath, sbPath, catEntry.size) 
        else:
            raise Exception("Non-CAS payloads are currently unsupported.")

    sb.close()


def lp(path):
    path = path.replace("/", "\\")
    if path[:4] == '\\\\?\\' or path == "" or len(path) <= 247: 
        return path
    
    return unicode('\\\\?\\' + os.path.normpath(path))


def open2(path, mode):
    path = path.replace("/", "\\")
    pathParts = path.split("\\")
    manualPart = "\\".join(pathParts[:2])

    if not os.path.isdir(manualPart): 
        os.makedirs(manualPart)

    folderPath = lp(os.path.dirname(path))
    
    if not os.path.isdir(folderPath): 
        os.makedirs(folderPath)
    
    return open(lp(path),mode)


def casPayload(bundleEntry, targetPath, sourcePath, size):
    catEntry = cat[bundleEntry.sha1]

    # Open the cas file so we can read the payload.
    cas = open(catEntry.path, "rb")
    cas.seek(catEntry.offset)

    # Open the output file so we can write the final data.
    out = open2(targetPath, "wb")

    leftSize = size
      
    while leftSize > 0:
        finalData = None

        # Read the data we need.
        finalSize, magic, payloadLen = unpack(">IHH", cas.read(8))
        leftSize -= 8

        if magic == 0x0F70:
            # Read & Decompress the data.
            payload = cas.read(payloadLen)
            finalData = zstd.decompress(payload)
            leftSize -= payloadLen
        elif magic == 0x0070 or magic == 0x0071:
            # Read already decompressed data.
            finalData = cas.read(finalSize)
            leftSize -= finalSize
        else:
            raise Exception("Unknown compressed block magic '%x'." % magic)

        out.write(finalData)
    
    cas.close()
    out.close()


def dumpRoot(root):
    for dir0, dirs, ff in os.walk(root):
        for fname in ff:
            if fname[-4:] == ".toc":
                fname = dir0 + "\\" + fname
                dump(fname, g_TargetDirectory)


def parseCatalogs(root, catDict):
    for dir0, dirs, ff in os.walk(root):
        for fname in ff:
            if fname[-4:] == ".cat":
                fname = dir0 + "\\" + fname
                readCat(catDict, fname, root)


# Make the paths absolute and normalize the slashes
for path in "g_MainCatalogPath", "g_MainDataRoot":
    if path in locals():
        locals()[path] = os.path.normpath(g_GameDirectory + "\\" + locals()[path])


g_TargetDirectory = os.path.normpath(g_TargetDirectory)

# Read the catalog files
cat = dict()

if "g_MainDataRoot" in locals(): 
    parseCatalogs(g_MainDataRoot, cat)
    dumpRoot(g_MainDataRoot)
