# DbObject parsing implementation by NoFaTe

from struct import unpack
from cStringIO import StringIO


def decode7Bit(f):
    tempSlice = unpack('B', f.read(1))[0]

    result = tempSlice & 0x7F
    shift = 7

    while ((tempSlice & 0x80) != 0):
        tempSlice = unpack('B', f.read(1))[0]
        result = result | ((tempSlice & 0x7F) << shift)
        shift = shift + 7

    return result


def decodeZigZag(f):
    v1 = decode7Bit(f)
    v2 = v1 >> 1
    v3 = v1 << 31
    return v2 ^ (v3 >> 31)


def readNullTerminatedString(f):
    result=""
    while 1:
        char=f.read(1)
        if char=="\x00": return result
        result+=char


class DbObjectType: 
    Eoo = 0x00
    InternalMin = 0x00
    Array = 0x01
    Object = 0x02
    HomoArray = 0x03
    Null = 0x04
    ObjectId = 0x05
    Bool = 0x06
    String = 0x07
    Integer = 0x08
    Long = 0x09
    VarInt = 0x0A
    Float = 0x0B
    Double = 0x0C
    Timestamp = 0x0D
    RecordId = 0x0E
    Guid = 0x0F
    Sha1 = 0x10
    Matrix44 = 0x11
    Vector4 = 0x12         
    Blob = 0x13
    Attachment = 0x14
    Timespan = 0x15
    InternalMax = 0x1F
    Mask = 0x1F
    Anonymous = 0x80
    
    def __init__(self):
        pass


class RecordId:
    ExtentId = 0
    PageId = 0
    SlotId = 0

    def __init__(self, reader):
        self.ExtentId = unpack('h', reader.read(2))[0]
        self.PageId = unpack('h', reader.read(2))[0]
        self.SlotId = unpack('h', reader.read(2))[0]


class DbObjectTimestamp:
    TimeData = 0

    def __init__(self, reader):
        self.TimeData = unpack('Q', reader.read(8))[0]
        # TODO: Properly parse this
  

class Vec4:
    X = 0
    Y = 0
    Z = 0
    W = 0

    def __init__(self, reader):
        self.X = unpack('f', reader.read(4))[0]
        self.Y = unpack('f', reader.read(4))[0]
        self.Z = unpack('f', reader.read(4))[0]
        self.W = unpack('f', reader.read(4))[0]


class Matrix44:
    XAxis = None
    YAxis = None
    ZAxis = None
    WAxis = None

    def __init__(self, reader):
        self.XAxis = Vec4(reader)
        self.YAxis = Vec4(reader)
        self.ZAxis = Vec4(reader)
        self.WAxis = Vec4(reader)

class DbObject:
    _elements = []

    def __init__(self, reader, length):
        self._elements = list()
        baseElement = None

        if length == -1:
            while (len(reader.read(1)) == 1):
                reader.seek(-1, 1)
                element = DbObjectElement(reader)

                if baseElement == None and element.ElementType == DbObjectType.Object and element.FieldName == '':
                    baseElement = element

                if element.ElementType == DbObjectType.Eoo:
                    continue

                if element.FieldName == '':
                    self._elements.append(element.Value)
                else:
                    vars(self)[element.FieldName] = element.Value

        else:
            endOffset = reader.tell() + length

            while (endOffset - reader.tell() > 0):
                element = DbObjectElement(reader)

                if baseElement == None and element.ElementType == DbObjectType.Object and element.FieldName == '':
                    baseElement = element

                if element.ElementType == DbObjectType.Eoo:
                    continue

                if element.FieldName == '':
                    self._elements.append(element.Value)
                else:
                    vars(self)[element.FieldName] = element.Value
      
    def get(self, field):
        return vars(self).get(field)


class DbObjectElement:
    ElementType = DbObjectType.Eoo
    FieldName = ''
    Value = None

    def __init__(self, reader):
        tempType = unpack('B', reader.read(1))[0]

        self.ElementType = (tempType & DbObjectType.Mask)

        if (self.ElementType == DbObjectType.Eoo) or ((tempType & DbObjectType.Anonymous)) != 0:
            self.FieldName = ''
        else:
            self.FieldName = readNullTerminatedString(reader)

        if self.ElementType == DbObjectType.Eoo or self.ElementType == DbObjectType.InternalMax or self.ElementType == DbObjectType.Null:
            return

        if self.ElementType == DbObjectType.Array:
            self.parseObject(reader)
            self.Value = self.Value._elements
            return

        if self.ElementType == DbObjectType.Object or self.ElementType == DbObjectType.HomoArray or self.ElementType == DbObjectType.ObjectId:
            self.parseObject(reader)
            return

        if self.ElementType == DbObjectType.Bool:
            self.parseBool(reader)
            return

        if self.ElementType == DbObjectType.String:
            self.parseString(reader)
            return

        if self.ElementType == DbObjectType.Integer:
            self.parseInteger(reader)
            return

        if self.ElementType == DbObjectType.Long:
            self.parseLong(reader)
            return

        if self.ElementType == DbObjectType.VarInt:
            self.parseVarInt(reader)
            return

        if self.ElementType == DbObjectType.Float:
            self.parseFloat(reader)
            return

        if self.ElementType == DbObjectType.Double:
            self.parseDouble(reader)
            return

        if self.ElementType == DbObjectType.Timestamp:
            self.parseTimestamp(reader)
            return

        if self.ElementType == DbObjectType.RecordId:
            self.parseRecordId(reader)
            return

        if self.ElementType == DbObjectType.Guid:
            self.parseGuid(reader)
            return

        if self.ElementType == DbObjectType.Sha1:
            self.parseSha1(reader)
            return

        if self.ElementType == DbObjectType.Blob:
            self.parseBlob(reader)
            return

        if self.ElementType == DbObjectType.Attachment:
            self.parseAttachment(reader)
            return

        if self.ElementType == DbObjectType.Timespan:
            self.parseTimespan(reader)
            return

        if self.ElementType == DbObjectType.Matrix44:
            self.parseMatrix44(reader)
            return

        raise Exception("Unknown DbObjectType encountered (%d)." % self.ElementType)

    def parseMatrix44(self, reader):
        self.Value = Matrix44(reader)

    def parseString(self, reader):
        length = decode7Bit(reader)
        self.Value = reader.read(length - 1)
        reader.read(1)

    def parseObject(self, reader):
        size = decode7Bit(reader)
        self.Value = DbObject(reader, size)

    def parseObjectId(self, reader):
        self.Value = reader.read(16) # Guid

    def parseBool(self, reader):
        self.Value = unpack('?', reader.read(1))[0]

    def parseInteger(self, reader):
        self.Value = unpack('i', reader.read(4))[0]

    def parseLong(self, reader):
        self.Value = unpack('q', reader.read(8))[0]

    def parseVarInt(self, reader):
        self.Value = decodeZigZag(reader)

    def parseFloat(self, reader):
        self.Value = unpack('f', reader.read(4))[0]

    def parseDouble(self, reader):
        self.Value = unpack('d', reader.read(8))[0]

    def parseTimestamp(self, reader):
        self.Value = DbObjectTimestamp(reader)

    def parseRecordId(self, reader):
        self.Value = RecordId(reader)

    def parseGuid(self, reader):
        self.Value = reader.read(16) # Guid

    def parseSha1(self, reader):
        self.Value = reader.read(20) # SHA1

    def parseBlob(self, reader):
        size = decode7Bit(reader)
        self.Value = reader.read(size)

    def parseAttachment(self, reader):
        self.Value = reader.read(20) # SHA1

    def parseTimespan(self, reader):
        self.Value = decodeZigZag(reader)

        
def UnXor(path):
    f = open(path,"rb")

    magic = unpack(">I", f.read(4))[0]

    if magic in (0x00D1CE00, 0x00D1CE01):
        f.seek(556)
        data = f.read()
    else:
        f.seek(0)
        data = f.read()

    f.close()

    return StringIO(data), len(data)


def FromFile(path):
    stream, length = UnXor(path)
    return FromStream(stream, length)


def FromStream(stream, length):
    return DbObject(stream, length)._elements[0]