# Original EBX dumping script by Frankelstner
# Fixed, updated and optimized by NoFaTe

from binascii import hexlify
import struct
import os
import cPickle
import copy
import multiprocessing
from multiprocessing import Manager, Process, Queue

# Adjust input and output folders here
inputFolder = r"H:\Research\FIFA17\Dump\Files\bundles\ebx"
outputFolder = r"M:\Research\FIFA17\Dump\EBX"
guidTableName = "guidTable fifa"  # Name of the guid table file; keeping separate names
# for separate games is highly recommended. The table is created at the location of the script.

EXTENSION = ".txt"  # Use a different file extension if you like.
SEP = "    "  # Adjust the amount of whitespace on the left of the converted file.

# Show offsets to the left
printOffsets = True  # True/False

# First run through all files to create a guid table to resolve external file references.
# Then run through all files once more, but this time convert them using the guid table.

guidTable = None


def main():
    if guidTable is None:
        createGuidTable()

    dumpText()

##############################################################
##############################################################

unpackLE = struct.unpack


def unpackBE(typ, data):
    return struct.unpack(">" + typ, data)


def fixguid(guid=''):
    g = guid.strip()

    if len(g) != 32:
        return g

    fixedguid = g[6:8] + g[4:6] + g[2:4] + g[0:2] + '-'
    fixedguid += g[10:12] + g[8:10] + '-'
    fixedguid += g[14:16] + g[12:14] + '-'
    fixedguid += g[16:20] + '-'
    fixedguid += g[20:]

    return fixedguid.upper()


def createGuidTable():
    global guidTable

    manager = Manager()
    ebxPaths = Queue()

    guidTable = manager.dict()

    for dir0, dirs, ff in os.walk(inputFolder):
        for fname in ff:
            if fname[-4:] != ".ebx":
                continue

            ebxPaths.put(dir0 + "\\" + fname)

    print 'The GUID table is being generated. Please wait, this can take a while...'

    processes = []

    for i in xrange(0, multiprocessing.cpu_count()):
        process = Process(target=processEBXGUID, args=(ebxPaths, guidTable))
        processes.append(process)
        process.start()

    for process in processes:
        process.join()

    guidTable = dict(guidTable)

    f5 = open(guidTableName, "wb")  # write the table
    cPickle.dump(guidTable, f5)
    f5.close()

    print 'GUID table generated and stored. Proceeding with deserialization.'


def processEBXGUID(paths, guidTbl):
    while not paths.empty():
        path = paths.get()

        f = open(lp(path), "rb")
        relPath = path[len(inputFolder):-4]

        if relPath[0] == "\\":
            relPath = relPath[1:]

        try:
            ebx = Ebx(f, relPath)
            guidTbl[ebx.fileGUID] = ebx.trueFilename
        except:
            pass

        f.close()


def dumpText():
    ebxPaths = Queue()

    for dir0, dirs, ff in os.walk(inputFolder):
        for fname in ff:
            if fname[-4:] != ".ebx":
                continue

            ebxPaths.put(dir0 + "\\" + fname)

    print 'Found %d EBX files to deserialize. Initializing %d processing threads...' \
          % (ebxPaths.qsize(), multiprocessing.cpu_count())

    print 'EBX files are now being deserialized and dumped. Please wait, this can take a while...'

    processes = []

    for i in xrange(0, multiprocessing.cpu_count()):
        process = Process(target=processEBX, args=(ebxPaths,))
        processes.append(process)
        process.start()

    for process in processes:
        process.join()

    print 'Finished deserializing and dumping EBX files.'


def processEBX(paths):
    while not paths.empty():
        path = paths.get()

        f = open(lp(path), "rb")
        relPath = path[len(inputFolder):-4]

        if relPath[0] == "\\":
            relPath = relPath[1:]

        try:
            ebx = Ebx(f, relPath)
            ebx.dump(outputFolder)
        except ValueError as msg:
            print 'An error occurred while deserializing "%s":\n%s' % (relPath, msg)

        f.close()


def open2(path, mode="rb"):
    path = path.replace("/", "\\")

    if mode == "wb":
        # create folders if necessary and return the file handle

        # first of all, create one folder level manually because makedirs might fail
        pathParts = path.split("\\")
        manualPart = "\\".join(pathParts[:2])

        if not os.path.isdir(manualPart):
            try:
                os.makedirs(manualPart)
            except:
                pass

        # now handle the rest, including extra long path names
        folderPath = lp(os.path.dirname(path))

        if not os.path.isdir(folderPath):
            try:
                os.makedirs(folderPath)
            except:
                pass

    return open(lp(path), mode)


def lp(path):  # long, normalized pathnames
    path = path.replace("/", "\\")
    if len(path) <= 247 or path == "" or path[:4] == '\\\\?\\': return os.path.normpath(path)
    return unicode('\\\\?\\' + os.path.normpath(path))


try:
    from ctypes import *

    floatlib = cdll.LoadLibrary("floattostring")

    def formatfloat(num):
        bufType = c_char * 100
        buf = bufType()
        bufpointer = pointer(buf)
        floatlib.convertNum(c_double(num), bufpointer, 100)
        rawstring = (buf.raw)[:buf.raw.find("\x00")]
        if rawstring[:2] == "-.":
            return "-0." + rawstring[2:]
        elif rawstring[0] == ".":
            return "0." + rawstring[1:]
        elif "e" not in rawstring and "." not in rawstring:
            return rawstring + ".0"
        return rawstring
except:
    def formatfloat(num):
        return str(num)


def hasher(keyword):  # 32bit FNV-1 hash with FNV_offset_basis = 5381 and FNV_prime = 33
    hash = 5381
    for byte in keyword:
        hash = (hash * 33) ^ ord(byte)
    return hash & 0xffffffff  # use & because Python promotes the num instead of intended overflow


class Header:
    def __init__(self, varList):
        self.absStringOffset = varList[0]  ## absolute offset for string section start
        self.lenStringToEOF = varList[1]  ## length from string section start to EOF
        self.numGUID = varList[2]  ## number of external GUIDs
        self.numInstanceRepeater = varList[3]  ## total number of instance repeaters
        self.numGUIDRepeater = varList[4]  ## instance repeaters with GUID
        self.unknown = varList[5]
        self.numComplex = varList[6]  ## number of complex entries
        self.numField = varList[7]  ## number of field entries
        self.lenName = varList[8]  ## length of name section including padding
        self.lenString = varList[9]  ## length of string section including padding
        self.numArrayRepeater = varList[10]
        self.lenPayload = varList[11]  ## length of normal payload section; the start of the array payload section is absStringOffset+lenString+lenPayload


class FieldDescriptor:
    def __init__(self, varList, keywordDict):
        self.name = keywordDict[varList[0]]
        self.type = varList[1]
        self.ref = varList[2]  # the field may contain another complex
        self.offset = varList[3]  # offset in payload section; relative to the complex containing it
        self.secondaryOffset = varList[4]
        if self.name == "$": self.offset -= 8

    def get_field_type(self):
        return (self.type >> 0x04) & 0x1F


class ComplexDescriptor:
    def __init__(self, varList, keywordDict):
        self.name = keywordDict[varList[0]]
        self.fieldStartIndex = varList[1]  # the index of the first field belonging to the complex
        self.numField = varList[2]  # the total number of fields belonging to the complex
        self.alignment = varList[3]
        self.type = varList[4]
        self.size = varList[5]  # total length of the complex in the payload section
        self.secondarySize = varList[6]  # seems deprecated


class InstanceRepeater:
    def __init__(self, varList):
        self.complexIndex = varList[0]  # index of complex used as the instance
        self.repetitions = varList[1]  # number of instance repetitions


class arrayRepeater:
    def __init__(self, varList):
        self.offset = varList[0]  # offset in array payload section
        self.repetitions = varList[1]  # number of array repetitions
        self.complexIndex = varList[2]  # not necessary for extraction


class Enumeration:
    def __init__(self):
        self.values = dict()
        self.type = 0


class Complex:
    def __init__(self, desc):
        self.desc = desc


class Field:
    def __init__(self, desc, offset):
        self.desc = desc
        self.offset = offset  # track absolute offset of each field in the ebx


class FieldType:
    Void = 0x0
    DbObject = 0x1
    ValueType = 0x2
    Class = 0x3
    Array = 0x4
    FixedArray = 0x5
    String = 0x6
    CString = 0x7
    Enum = 0x8
    FileRef = 0x9
    Boolean = 0xA
    Int8 = 0xB
    UInt8 = 0xC
    Int16 = 0xD
    UInt16 = 0xE
    Int32 = 0xF
    UInt32 = 0x10
    Int64 = 0x11
    UInt64 = 0x12
    Float32 = 0x13
    Float64 = 0x14
    Guid = 0x15
    SHA1 = 0x16
    ResourceRef = 0x17
    
    def __init__(self):
        pass


class Ebx:
    def __init__(self, f, relPath):
        # metadata
        magic = f.read(4)
        if magic == "\xCE\xD1\xB2\x0F":
            self.unpack = unpackLE
        elif magic == "\x0F\xB2\xD1\xCE":
            self.unpack = unpackBE
        else:
            raise ValueError("The file is not ebx: " + relPath)

        self.relPath = relPath  # to give more feedback for unknown field types
        self.trueFilename = ""
        self.header = Header(self.unpack("3I6H3I", f.read(36)))
        self.arraySectionstart = self.header.absStringOffset + self.header.lenString + self.header.lenPayload
        self.fileGUID = f.read(16)

        while f.tell() % 16 != 0:
            f.seek(1, 1)  # padding

        self.externalGUIDs = [(f.read(16), f.read(16)) for i in xrange(self.header.numGUID)]
        self.keywords = str.split(f.read(self.header.lenName), "\x00")
        self.keywordDict = dict((hasher(keyword), keyword) for keyword in self.keywords)

        self.fieldDescriptors = [FieldDescriptor(self.unpack("IHHii", f.read(16)), self.keywordDict) for i in
                                 xrange(self.header.numField)]

        self.complexDescriptors = [ComplexDescriptor(self.unpack("IIBBHHH", f.read(16)), self.keywordDict) for i in
                                   xrange(self.header.numComplex)]

        self.instanceRepeaters = [InstanceRepeater(self.unpack("2H", f.read(4))) for i in
                                  xrange(self.header.numInstanceRepeater)]

        self.enumerations = dict()

        while f.tell() % 16 != 0:
            f.seek(1, 1)  # padding

        self.arrayRepeaters = [arrayRepeater(self.unpack("3I", f.read(12))) for i in
                               xrange(self.header.numArrayRepeater)]

        # payload
        f.seek(self.header.absStringOffset + self.header.lenString)
        self.internalGUIDs = []
        self.instances = []  # (guid, complex)
        nonGUIDindex = 0
        self.isPrimaryInstance = True  # first instance is primary
        for i, instanceRepeater in enumerate(self.instanceRepeaters):
            for repetition in xrange(instanceRepeater.repetitions):
                # obey alignment of the instance; peek into the complex for that
                while f.tell() % self.complexDescriptors[instanceRepeater.complexIndex].alignment != 0: f.seek(1, 1)

                # all instances after numGUIDRepeater have no guid
                if i < self.header.numGUIDRepeater:
                    instanceGUID = f.read(16)
                else:
                    # just numerate those instances without guid and assign a big endian int to them.
                    instanceGUID = struct.pack(">I", nonGUIDindex)
                    nonGUIDindex += 1
                self.internalGUIDs.append(instanceGUID)

                self.instances.append((instanceGUID, self.readComplex(instanceRepeater.complexIndex, f, True)))
                self.isPrimaryInstance = False  # the readComplex function has used isPrimaryInstance by now
        f.close()

        # if no filename found, use the relative input path instead
        # it's just as good though without capitalization
        if self.trueFilename == "":
            self.trueFilename = relPath

    def readComplex(self, complexIndex, f, isInstance=False):
        complexDesc = self.complexDescriptors[complexIndex]
        cmplx = Complex(complexDesc)
        cmplx.offset = f.tell()

        cmplx.fields = []
        # alignment 4 instances require subtracting 8 for all field offsets and the complex size
        obfuscationShift = 8 if (isInstance and cmplx.desc.alignment == 4) else 0

        for fieldIndex in xrange(complexDesc.fieldStartIndex, complexDesc.fieldStartIndex + complexDesc.numField):
            f.seek(cmplx.offset + self.fieldDescriptors[fieldIndex].offset - obfuscationShift)
            cmplx.fields.append(self.readField(fieldIndex, f))

        f.seek(cmplx.offset + complexDesc.size - obfuscationShift)
        return cmplx

    def readField(self, fieldIndex, f):
        fieldDesc = self.fieldDescriptors[fieldIndex]

        field = Field(fieldDesc, f.tell())

        if fieldDesc.get_field_type() == FieldType.Void:
            # Void (inheritance)
            field.value = self.readComplex(fieldDesc.ref, f)

        elif fieldDesc.get_field_type() == FieldType.ValueType:
            # ValueType
            field.value = self.readComplex(fieldDesc.ref, f)

        elif fieldDesc.get_field_type() == FieldType.Class:
            # Class (reference)
            field.value = self.unpack('I', f.read(4))[0]

        elif fieldDesc.get_field_type() == FieldType.Array:
            # Array
            array_repeater = self.arrayRepeaters[self.unpack("I", f.read(4))[0]]
            array_complex_desc = self.complexDescriptors[fieldDesc.ref]

            f.seek(self.arraySectionstart + array_repeater.offset)
            array_complex = Complex(array_complex_desc)
            array_complex.fields = [self.readField(array_complex_desc.fieldStartIndex, f) for repetition in
                                    xrange(array_repeater.repetitions)]
            field.value = array_complex

        elif fieldDesc.get_field_type() == FieldType.CString:
            # CString
            startPos = f.tell()
            stringOffset = self.unpack("i", f.read(4))[0]
            if stringOffset == -1:
                field.value = "*nullString*"
            else:
                f.seek(self.header.absStringOffset + stringOffset)
                field.value = ""
                while 1:
                    a = f.read(1)
                    if a == "\x00":
                        break
                    else:
                        field.value += a
                f.seek(startPos + 4)

                if self.isPrimaryInstance and fieldDesc.name == "Name" and self.trueFilename == "":
                    self.trueFilename = field.value

        elif fieldDesc.get_field_type() == FieldType.Enum:
            # Enum
            compareValue = self.unpack("i", f.read(4))[0]
            enumComplex = self.complexDescriptors[fieldDesc.ref]

            if fieldDesc.ref not in self.enumerations:
                enumeration = Enumeration()
                enumeration.type = fieldDesc.ref

                for i in xrange(enumComplex.fieldStartIndex, enumComplex.fieldStartIndex + enumComplex.numField):
                    enumeration.values[self.fieldDescriptors[i].offset] = self.fieldDescriptors[i].name

                self.enumerations[fieldDesc.ref] = enumeration

            if compareValue not in self.enumerations[fieldDesc.ref].values:
                field.value = '*nullEnum*'
            else:
                field.value = self.enumerations[fieldDesc.ref].values[compareValue]

        elif fieldDesc.get_field_type() == FieldType.FileRef:
            # FileRef
            startPos = f.tell()
            stringOffset = self.unpack("i", f.read(4))[0]
            if stringOffset == -1:
                field.value = "*nullString*"
            else:
                f.seek(self.header.absStringOffset + stringOffset)
                field.value = ""
                while 1:
                    a = f.read(1)
                    if a == "\x00":
                        break
                    else:
                        field.value += a
                f.seek(startPos + 4)

                if self.isPrimaryInstance and fieldDesc.name == "Name" and self.trueFilename == "":
                    self.trueFilename = field.value

        elif fieldDesc.get_field_type() == FieldType.Boolean:
            # Boolean
            field.value = self.unpack('?', f.read(1))[0]

        elif fieldDesc.get_field_type() == FieldType.Int8:
            # Int8
            field.value = self.unpack('c', f.read(1))[0]

        elif fieldDesc.get_field_type() == FieldType.UInt8:
            # UInt8
            field.value = self.unpack('B', f.read(1))[0]

        elif fieldDesc.get_field_type() == FieldType.Int16:
            # Int16
            field.value = self.unpack('h', f.read(2))[0]

        elif fieldDesc.get_field_type() == FieldType.UInt16:
            # UInt16
            field.value = self.unpack('H', f.read(2))[0]

        elif fieldDesc.get_field_type() == FieldType.Int32:
            # Int32
            field.value = self.unpack('i', f.read(4))[0]

        elif fieldDesc.get_field_type() == FieldType.UInt32:
            # UInt32
            field.value = self.unpack('I', f.read(4))[0]

        elif fieldDesc.get_field_type() == FieldType.Int64:
            # Int64
            field.value = self.unpack('q', f.read(8))[0]

        elif fieldDesc.get_field_type() == FieldType.UInt64:
            # UInt64
            field.value = self.unpack('Q', f.read(8))[0]

        elif fieldDesc.get_field_type() == FieldType.Float32:
            # Float32
            field.value = self.unpack('f', f.read(4))[0]

        elif fieldDesc.get_field_type() == FieldType.Float64:
            # Float64
            field.value = self.unpack('d', f.read(8))[0]

        elif fieldDesc.get_field_type() == FieldType.Guid:
            # Guid
            field.value = f.read(16)

        elif fieldDesc.get_field_type() == FieldType.SHA1:
            # SHA1
            field.value = f.read(20)

        elif fieldDesc.get_field_type() == FieldType.ResourceRef:
            # ResourceRef
            field.value = f.read(8)

        else:
            # Unknown
            print 'Unknown field type "%d"' % fieldDesc.get_field_type()

        return field

    def dump(self, outFolder):
        # if not self.trueFilename: self.trueFilename=hexlify(self.fileGUID)

        outName = outFolder + self.trueFilename + EXTENSION

        # dirName=os.path.dirname(outputFolder+self.trueFilename)
        # if not os.path.isdir(dirName): os.makedirs(dirName)
        # if not self.trueFilename: self.trueFilename=hexlify(self.fileGUID)
        # f2=open(outputFolder+self.trueFilename+EXTENSION,"wb")
        f2 = open2(outName, "wb")

        for (guid, instance) in self.instances:
            # print
            self.writeInstance(f2, instance, fixguid(hexlify(guid)))
            self.recurse(instance.fields, f2, 0)
        f2.close()

    def writeField(self, f, field, lvl, text):
        if printOffsets:
            f.write(hex2(field.offset) + SEP + lvl * SEP + field.desc.name + text + "\r\n")
        else:
            f.write(lvl * SEP + field.desc.name + text + "\r\n")

    def writeInstance(self, f, cmplx, text):
        if printOffsets:
            f.write(hex2(cmplx.offset) + SEP + cmplx.desc.name + " " + text + "\r\n")
        else:
            f.write(cmplx.desc.name + " " + text + "\r\n")

    def recurse(self, fields, f2, lvl):  # over fields
        lvl += 1
        for field in fields:
            if field.desc.get_field_type() == FieldType.Void or field.desc.get_field_type() == FieldType.ValueType:
                self.writeField(f2, field, lvl, "::" + field.value.desc.name)
                self.recurse(field.value.fields, f2, lvl)

            elif field.desc.get_field_type() == FieldType.Class:
                if field.value >> 31:
                    extguid = self.externalGUIDs[field.value & 0x7fffffff]

                    try:
                        towrite = guidTable[extguid[0]] + "/" + fixguid(hexlify(extguid[1]))
                    except:
                        towrite = fixguid(hexlify(extguid[0])) + "/" + fixguid(hexlify(extguid[1]))

                elif field.value == 0:
                    towrite = "*nullGuid*"

                else:
                    intGuid = self.internalGUIDs[field.value - 1]
                    towrite = fixguid(hexlify(intGuid))

                self.writeField(f2, field, lvl, " " + towrite)

            elif field.desc.get_field_type() == FieldType.Array:
                if len(field.value.fields) == 0:
                    self.writeField(f2, field, lvl, " *nullArray*")
                else:
                    self.writeField(f2, field, lvl, "::" + field.value.desc.name)

                    # quick hack so I can add indices to array members while using the same recurse function
                    for index in xrange(len(field.value.fields)):
                        member = field.value.fields[index]

                        if member.desc.name == "member":
                            desc = copy.deepcopy(member.desc)
                            desc.name = "member(" + str(index) + ")"
                            member.desc = desc
                            field.value.fields[index] = member

                    self.recurse(field.value.fields, f2, lvl)

            elif field.desc.get_field_type() == FieldType.Guid:
                self.writeField(f2, field, lvl, " " + fixguid(hexlify(field.value)))

            elif field.desc.get_field_type() == FieldType.SHA1:
                self.writeField(f2, field, lvl, " " + hexlify(field.value).upper())

            elif field.desc.get_field_type() == FieldType.ResourceRef:
                val = hexlify(field.value[::-1])
                # val=val[:16]+"/"+val[16:]
                self.writeField(f2, field, lvl, " " + val)

            else:
                self.writeField(f2, field, lvl, " " + str(field.value))


def hex2(num):
    # take int, return 8byte string
    a = hex(num)
    if a[:2] == "0x": a = a[2:]
    if a[-1] == "L": a = a[:-1]
    while len(a) < 8:
        a = "0" + a
    return a

if outputFolder[-1] not in ("/", "\\"):
    outputFolder += "\\"

if inputFolder[-1] not in ("/", "\\"):
    inputFolder += "\\"

# if there's a guid table already, use it
try:
    f5 = open(guidTableName, "rb")
    guidTable = cPickle.load(f5)
    f5.close()
except:
    guidTable = None

if __name__ == '__main__':
    main()
